def create_interval(string: str):
    bounds = string.split("_")[-1].replace("[", "").replace(")", "").split(",")
    return float(bounds[0]), float(bounds[1])


def add_column_expectations(batch, col_profiles):
    """
    Returns the sum of two decimal numbers in binary digits.

            Parameters:
                    batch (int): A decimal integer
                    column_types
                    col_profiles (int): Another decimal integer

            Returns:
                    binary_sum (str): Binary string of the sum of a and b
    """
    for i in col_profiles:
        if col_profiles[i]["type"] == "number" and col_profiles[i]["na_frac"] < 1.0:

            batch.expect_column_values_to_be_between(
                i,
                min_value=col_profiles[i]["min"],
                max_value=col_profiles[i]["max"],
            )
            batch.expect_column_mean_to_be_between(
                i,
                col_profiles[i]["mean"] - col_profiles[i]["mean"] * 0.2,
                col_profiles[i]["mean"] + col_profiles[i]["mean"] * 0.2,
            )
            batch.expect_column_stdev_to_be_between(
                i,
                col_profiles[i]["std"] - col_profiles[i]["std"] * 0.3,
                col_profiles[i]["std"] + col_profiles[i]["std"] * 0.3,
            )
        elif (
            col_profiles[i]["type"] == "string"
            and col_profiles[i]["na_frac"] < 1.0
            and col_profiles[i]["n_unique"] > 1
        ):
            batch.expect_column_most_common_value_to_be_in_set(
                i,
                col_profiles[i]["most_frequent"],
            )

    return batch


def add_conditional_expectations(batch, test_list, prepro, rare_labels):

    rl_dict = {}
    for i in rare_labels:
        rl_dict[i[0]] = list(i[1])

    for test in test_list:

        feat1 = [i for i in prepro.cat_cols.keys() if test[0] in prepro.cat_cols[i]][0]
        feat2 = [i for i in prepro.cat_cols.keys() if test[1] in prepro.cat_cols[i]][0]

        if str(batch[feat2].dtype) not in ["object", "bool"]:
            interval = create_interval(test[1])
            parse_arg = (
                "`"
                + feat2
                + "`>"
                + str(interval[0])
                + " and "
                + "`"
                + feat2
                + "`<"
                + str(interval[1])
            )
        else:
            value2 = test[1].replace(feat2 + "_", "")
            parse_arg = "`" + feat2 + '`=="' + value2 + '"'

        if str(batch[feat1].dtype) not in ["object", "bool"]:
            interval = create_interval(test[0])
            batch.expect_column_values_to_be_between(
                column=feat1,
                min_value=interval[0],
                max_value=interval[1],
                condition_parser="pandas",
                row_condition=parse_arg,
                mostly=test[2],
            )
        else:
            value1 = test[0].replace(feat1 + "_", "")
            if value1 == "Grouped_labels":
                set1 = rl_dict[feat1]
            else:
                set1 = [value1]

            batch.expect_column_values_to_be_in_set(
                column=feat1,
                value_set=set1,
                condition_parser="pandas",
                row_condition=parse_arg,
                mostly=test[2],
            )

    return batch


def column_greater_than(data_batch, column_list):

    for test_i in column_list:
        data_batch.expect_column_pair_values_A_to_be_greater_than_B(
            column_A=test_i[0],
            column_B=test_i[1],
        )

    return data_batch

    # TO DO
    # e2 = batch.expect_column_pair_values_A_to_be_greater_than_B('loan_amount',
    # 'fico_average')
    # expect_column_pair_values_to_be_in_set
    # expect_column_values_to_be_dateutil_parseable
    # expect_column_values_to_be_null
    # expect_column_values_to_be_unique
    # expect_column_values_to_be_valid_date
    # expect_column_values_to_match_strftime_format
    # expect_multicolumn_sum_to_equal



import pickle
from typing import Dict

import pandas as pd

from .profiler import DatasetProfiler


class MultiTableProfiler:
    """

    The DatasetProfiler class .

    """

    def __init__(
        self,
        relational_metadata: Dict,
        target: Dict = None,
        n_samples: int = 10000,
        compression: str = None,
        separator: str = ",",
        thousands: str = None,
        decimals: str = ".",
        encoding: str = None,
    ):
        """

        Parameters
        ----------s
        df_path : str
            path of the CSV file to be profiled.
        primary_key : str, optional
            name of the column defining the CSV primary key (composite).
        sequence_index : str, optional
            name of the column from the CSV containing a sequence index.
        target : str, optional
            name of the column from the CSV containing a supervised learning target
            variable.
        compression : :obj:`int`, optional
            Description of `param3`.

        """
        self.profiles = []
        self.relational_metadata = relational_metadata
        self.io_meta = {
            "compression": compression,
            "separator": separator,
            "thousands": thousands,
            "decimals": decimals,
            "encoding": encoding,
        }

        tables = {}
        try:
            for i in relational_metadata.keys():
                # print(i, len(relational_metadata[i]))
                table_i = pd.read_csv(
                    relational_metadata[i][0],
                    compression=compression,
                    sep=separator,
                    decimal=decimals,
                    thousands=thousands,
                )

                table_i.index = table_i[relational_metadata[i][1]]
                table_i = table_i.drop(relational_metadata[i][1], axis=1)
                tables[i] = table_i
            # check if primary key exists
            # assign primary key to index
            # check if parent table exists
            #
        except FileNotFoundError:
            print("Could not read tables provided as paths")
            return

        completed = []
        to_process = list(relational_metadata.keys())
        w = 0
        while len(to_process) > 0 and w < 100:

            for i in to_process:
                len_meta = len(relational_metadata[i]) / 2
                if len_meta == 1:
                    completed.append(i)
                    to_process.remove(i)
                else:
                    foreign_dfs = [
                        relational_metadata[i][2 + 2 * j]
                        for j in range(int(len_meta) - 1)
                    ]
                    if all(elem in completed for elem in foreign_dfs):
                        completed.append(i)
                        to_process.remove(i)
            w += 1

        if w > 99:
            print("Could not build relational DAG, please check foreign keys provided")
            return

        metadata_merging = {}
        foreign_keys = {}

        for i in completed:
            len_meta = len(relational_metadata[i]) / 2
            if len_meta > 1:
                df = tables[i]
                columns = list(df.columns)
                for j in range(int(len_meta) - 1):
                    df = df.join(
                        tables[relational_metadata[i][2 + 2 * j]],
                        # lsuffix='_l_'+i,
                        rsuffix="_r_" + relational_metadata[i][2 + 2 * j],
                        on=relational_metadata[i][3 + 2 * j],
                    )
                    df = df.drop(relational_metadata[i][3 + 2 * j], axis=1)

                tables[i] = df.sample(min(n_samples, df.shape[0]))
                metadata_merging[i] = columns
                # foreign_keys[i] = relational_metadata[i][3 + 2 * i]
                # print(i, df.shape)
            else:
                metadata_merging[i] = []

        self.tables = tables
        self.metadata = metadata_merging
        self.foreign_keys = foreign_keys

        return

    def profile(self, n_samples=10000):

        for i in self.tables.keys():
            self.tables[i].to_csv("temp.csv", index_label="index")
            dp = DatasetProfiler(
                "temp.csv",
                primary_key="index",
                # regression=self.io_meta['regression'],
                n_samples=n_samples,
                # compression=self.io_meta['compression'],
                # separator=self.io_meta['separator'],
                # decimals=self.io_meta['decimals'],
                # thousands=self.io_meta['thousands'],
                # encoding=self.io_meta['encoding']
            )

            dp.profile()
            # dp.generate_expectations(suite_name=str(i))

            dp.data_sample = []
            dp.reduced_data_sample = []

            self.profiles.append(dp)

        print("Profiling finished.")

        return

    def save(self, name: str):

        with open(name, "wb") as f:
            pickle.dump(self, f)



import copy
import itertools
import re

from distfit import distfit
import numpy as np
import pandas as pd
import scipy.stats as ss
from sklearn.linear_model import LinearRegression
from sklearn.tree import DecisionTreeClassifier  # ,export_text
from tqdm import tqdm


def fit_distributions(x: pd.Series):

    if 10 > x.nunique() > 1:
        dist = distfit(method="discrete")
        dist.fit_transform(x.fillna(0), verbose=False)
        out = [dist.model["name"], dist.model["score"]]
    elif x.nunique() >= 10:
        dist = distfit()
        dist.fit_transform(x.fillna(0), verbose=False)
        out = [dist.model["name"], dist.model["score"]]

    return out


def check_precision(x, tol=1e-8):

    types = ["int16", "int32", "float16", "float32"]
    l1 = (x.fillna(0).astype(np.int16) - x.fillna(0)).mean()
    l2 = (x.fillna(0).astype(np.int32) - x.fillna(0)).mean()
    l3 = (x.fillna(0).astype(np.float16) - x.fillna(0)).mean()
    l4 = (x.fillna(0).astype(np.float32) - x.fillna(0)).mean()

    residuals = np.abs([l1, l2, l3, l4]) < tol
    if residuals.sum() > 0:
        val = next((index for index, value in enumerate(residuals) if value != 0), None)
        prec = types[val]
    else:
        prec = "float64"

    return prec


def find_deterministic_columns_binary(df, binary):
    jtr = np.random.choice(np.arange(df.shape[0]), 5000, replace=True)
    jts = np.random.choice(np.arange(df.shape[0]), 2000, replace=True)
    deterministic = []
    numerical_cols = [
        i for i in df.columns if df[i].dtype != "object" if i not in binary
    ]
    for j, i in enumerate(binary):
        y = df[i].astype("category").cat.codes
        clf = DecisionTreeClassifier(max_depth=1).fit(
            df[numerical_cols].fillna(0).iloc[jtr],
            y.iloc[jtr].astype(np.int32),
        )
        score = clf.score(
            df[numerical_cols].fillna(0).iloc[jts],
            y.iloc[jts].astype(np.int32),
        )

        # text = export_text(clf, feature_names=list(df[numerical_cols].columns))
        # print(export_text(clf, decimals=4))
        # print(i, text, score)
        #
        # print(float(text.split('|')[-3].split('>')[-1]))
        # print(float(text.split('|')[-1].split(':')[-1]))

        if score > 0.999:
            deterministic.append([i, clf, copy.deepcopy(numerical_cols)])

    return deterministic, numerical_cols


def find_deterministic_columns_regression(df):
    jtr = np.random.choice(np.arange(df.shape[0]), 1000, replace=True)  # TOFIX
    jts = np.random.choice(np.arange(df.shape[0]), 100, replace=True)

    deterministic_num = []
    numerical_cols = [i for i in df.columns if df[i].dtype != "object"]
    if len(numerical_cols) <= 1:
        return deterministic_num, numerical_cols

    for j, i in enumerate(tqdm(numerical_cols)):
        clf = LinearRegression(n_jobs=-1, positive=True, fit_intercept=True).fit(
            df[numerical_cols].fillna(0).drop(i, axis=1).iloc[jtr],
            df[i].fillna(0).iloc[jtr],
        )

        score = clf.score(
            df[numerical_cols].fillna(0).drop(i, axis=1).iloc[jts],
            df[i].fillna(0).iloc[jts],
        )
        if score > 0.9999:
            numerical_cols.remove(i)
            deterministic_num.append([i, clf, copy.deepcopy(numerical_cols)])

    return deterministic_num, numerical_cols


def find_ordinal_columns(df, cat_columns):
    ordinal_dicts = {}

    for i in tqdm(cat_columns):
        if 2 < df[i].nunique() < 255:
            s = []
            unique = df[i].fillna("N/A").unique()
            for j in unique:
                lp = np.array([int(i) for i in re.findall(r"\d+", j.replace(".", ""))])
                if len(lp) == 0:
                    lp = np.array([-1])

                s.append((lp.max() + lp.min()) / 2)

            possible_dict = dict(
                zip([unique[i] for i in np.argsort(s)], np.arange(unique.shape[0])),
            )

            if (
                np.unique(list(possible_dict.values())).shape[0]
                / len(list(possible_dict.values()))
                == 1
            ) and (np.mean(s) > 0):
                ordinal_dicts[i] = possible_dict

    return ordinal_dicts


def get_features_correlation(X):
    def _cramers_corrected_stat(confusion_matrix):
        """
        Calculate Cramers V statistic for categorial-categorial association,
        using correction from Bergsma and Wicher Journal of the Korean
        Statistical Society 42 (2013): 323-328
        """

        chi2 = ss.chi2_contingency(confusion_matrix)[0]
        n = confusion_matrix.sum().sum()
        phi2 = chi2 / n
        r, k = confusion_matrix.shape
        phi2corr = max(0, phi2 - ((k - 1) * (r - 1)) / (n - 1))
        rcorr = r - ((r - 1) ** 2) / (n - 1)
        kcorr = k - ((k - 1) ** 2) / (n - 1)
        return np.sqrt(phi2corr / min((kcorr - 1), (rcorr - 1)))

    features_correlation = pd.DataFrame(index=X.columns, columns=X.columns)
    print("Calculating correlation matrix:")
    Xsample = copy.deepcopy(X)
    if X.shape[1] > 100:
        n = 500
    else:
        n = 2000
    samples = min(n, X.shape[0])
    Xsample = Xsample.sample(n=samples)
    for feature_i in tqdm(Xsample.columns):
        for feature_j in Xsample.columns:
            confusion_matrix = pd.crosstab(
                Xsample[feature_i].fillna(0),
                Xsample[feature_j].fillna(0),
            )

            features_correlation[feature_i].loc[feature_j] = float(
                round(_cramers_corrected_stat(confusion_matrix), 4),
            )

    return features_correlation


def get_label_correlation(
    Xproc,
    cat_cols,
    p_tr=0.75,
    delta_tr=0.05,
    n_min=100,
):
    list2d = list(cat_cols.values())
    merged = list(itertools.chain(*list2d))
    corr = []
    Xnp = Xproc.values
    for j in tqdm(range(len(merged))):
        Xt = Xnp[Xnp[:, j] == 1]
        d = Xt.shape[0]
        sample2 = np.random.choice(np.arange(d), d, replace=False)
        for i in range(len(merged)):
            p = Xt[:, i].sum() / d
            p2 = Xnp[sample2, i][sample2].sum() / d
            delta = p - p2
            if p > p_tr and i != j and d > n_min and delta > delta_tr:
                corr.append((merged[i], merged[j], p, d / Xproc.shape[0]))

    return corr


def column_a_greater_than_b(x, column_types, t=1.0):
    num_cols = [i for i in x.columns if column_types[i] == "number"]
    comp_matrix = dict()
    for i in tqdm(num_cols):
        for j in num_cols:
            d = (x[i] - x[j]).dropna()
            std = 1.0 * x[i].std()
            p = (d >= 0).sum() / d.shape[0]
            d_range = (np.abs(d) < std).sum() / d.shape[0]
            if i != j and p >= t and d_range >= t and d.shape[0] / x.shape[0] > 0.1:
                comp_matrix[i, j] = (d >= 0).sum() / d.shape[0]

    return comp_matrix



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['structured_data_profiling',
 'structured_data_profiling.data_slicing',
 'structured_data_profiling.data_tests',
 'structured_data_profiling.expectations',
 'structured_data_profiling.preprocessor',
 'structured_data_profiling.profiler',
 'structured_data_profiling.utils']

package_data = \
{'': ['*']}

install_requires = \
['dateparser>=1,<2',
 'distfit>=1.4,<2.0',
 'great-expectations>=0.15.21,<0.16.0',
 'numpy>=1.21,<2.0',
 'pandas>=1.3.0,<2.0.0',
 'scikit-learn>=1.0,<2.0',
 'scipy>=1.7.2,<1.8.0']

setup_kwargs = {
    'name': 'structured-profiling',
    'version': '0.3.5',
    'description': 'A Python library to check for data quality and automatically generate data tests. ',
    'long_description': '\n## StructuredDataProfiling\n\nThe StructuredDataProfiling is a Python library developed to automatically profile structured datasets and to facilitate the creation of **data tests**.\n\nThe library creates data tests in the form of **Expectations** using the [great_expectations](https://www.greatexpectations.io) framework. Expectations are \'declarative statements that a computer can evaluate and semantically meaningful to humans\'.\n\nAn expectation could be, for example, \'the sum of columns a and b should be equal to one\' or \'the values in column c should be non-negative\'.\n\nStructuredDataProfiling runs a series of tests aimed at identifying statistics, rules, and constraints characterising a given dataset. The information generated by the profiler is collected by performing the following operations:\n\n- Characterise uni- and bi-variate distributions.\n- Identify data quality issues.\n- Evaluate relationships between attributes (ex. column C is the difference between columns A and B)\n- Understand ontologies characterizing categorical data (column A contains names, while B contains geographical places).\n\nFor an overview of the library outputs please check the [examples](./examples) section.\n\n# Installation\nYou can install StructuredDataProfiling by using pip:\n`pip install structured-profiling\n`\n# Quickstart\nYou can import the profiler using\n\n```python\nfrom structured_data_profiling.profiler import DatasetProfiler\n```\nYou can import the profiler using\n```python\nprofiler = DatasetProfiler(\'./csv_path.csv\')\n```\nThe presence of a primary key (for example to define relations between tables or sequences) can be specified by using the argument **primary key** containing a single or multiple column names.\n\nTo start the profiling scripts, you can run the profile() method\n```python\nprofiler.profile()\n```\nThe method generate_expectations() outputs the results of the profiling process converted into data expectations. Please note, the method requires the existence of a local great_expectations project.\nIf you haven\'t done so please run ```great_expectations init``` in your working directory.\n```python\nprofiler.generate_expectations()\n```\nThe expectations are generated in a JSON format using the great_expectation schema. The method will also create data docs using the rendered provided by the great_expectations library.\n\nThese docs can be found in the local folder ```great_expectations/uncommitted/data_docs```.\n\n# Profiling outputs\nThe profiler generates 3 json files describing the ingested dataset. These json files contain information about:\n- column_profiles: it contains the statistical characterisation of the dataset columns.\n- dataset_profile: it highlights issues and limitations affecting the dataset.\n- tests: it contains the data tests found by the profiler.\n\nThe process of generating expectations makes use of the great_expectations library to produce an HTML file contaning data docs. An example of data doc for a given column can be seen in the image below.\n\n<img alt="data docs example 1" src="https://raw.githubusercontent.com/Clearbox-AI/StructuredDataProfiling/main/examples/num_columns.PNG"/>\n\n\n# Examples\nYou can find a couple of notebook examples in the [examples](./examples) folder.\n# To-dos\nDisclaimer: this library is still at a very early stage. Among other things, we still need to:\n\n- [ ] Support more data formats (Feather, Parquet)\n- [ ] Add more Expectations\n- [ ] Integrate PII identification using Presidio\n- [ ] Optimise and compile part of the profiling routines using Cython\n- [ ] Write library tests\n',
    'author': 'Clearbox AI',
    'author_email': 'info@clearbox.ai',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Clearbox-AI/StructuredDataProfiling',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.1,<3.11',
}


setup(**setup_kwargs)
